class DatabaseAccessError(Exception):
    pass


class DatabaseEntrySpecificationError(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.message}"


class DatabaseEntryMissingColumnError(Exception):
    def __init__(self, cls, column):
        self.cls = cls
        self.column = column

    def __repr__(self):
        return f"{self.__class__.__name__}: Class {self.cls.__name__} perform this operation. This functionality requires a database with column '{self.column}'"


class InvalidTagValueError(Exception):
    def __init__(self, tag):
        self.tag = tag

    def __repr__(self):
        return f"{self.__class__.__name__}: The tag value: '{self.tag}' could not be processed."


class WebResourceError(Exception):
    pass


class CatalogResourceError(WebResourceError):
    pass



import logging

log = logging.getLogger()

try:
    import minimock

    mock = minimock.mock
except:
    log.warning("Unable to import minimock")


try:
    import doctest
except:
    log.warning("Unable to import doctest")



from pathlib import Path
import logging

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import OperationalError

from forgefrenzy.exceptions import *
from forgefrenzy.blueplate.logger import log

logging.getLogger("sqlalchemy.engine").setLevel(logging.INFO)

EntryORM = declarative_base()


class DFDB:
    def __init__(self, volatile=False, echo=False):
        root = Path(__file__).parent
        database = root / "df.sqlite"

        if volatile:
            log.warning("Starting database in volatile memory")
            self.db = "sqlite:///:memory:"
        else:
            self.db = f"sqlite:///{database}"

        self.engine = create_engine(self.db, echo=echo)

        self.Session = sessionmaker(bind=self.engine)

    def __str__(self):
        return f"DFDB at <{self.db}>"

    def generate_table(self, table_class):
        table_class.metadata.create_all(self.engine)

    @property
    def session(self):
        # return self.Session.begin()
        return DatabaseSession(self)


dfdb = DFDB()


class DatabaseSession:
    """Simple wrapper around existing session context manager"""

    def __init__(self, db):
        self.debug = False
        self.session = db.Session()

    def __enter__(self):
        if self.debug:
            log.info(f"Entering session {self.session}")
        return self.session.__enter__()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.debug:
            log.info(f"Exiting session {self.session}")

        return self.session.__exit__(exc_type, exc_value, exc_traceback)


class DatabaseEntry:
    db = dfdb
    __table__ = None
    __primary__ = None
    tagstring = None

    def __init__(self, **kwargs):
        log.debug(f"Initializing object from values")
        if isinstance(self, EntryORM):
            EntryORM.__init__(self, **kwargs)

        log.debug(f"Created {self} -> {self.db}")

    def __repr__(self):
        return f"{self.__class__.__name__} <{self.primary}>"

    def __str__(self):
        return f"{self.__class__.__name__} <{self.primary}>"

    @property
    def primary(self):
        """
        Returns the value of the primary key for this object
        """
        return self.__getattribute__(self.__primary__)

    @property
    def session(self):
        """Returns a database session"""
        return DatabaseSession(self.db)

    @property
    def as_dict(self):
        """Returns the object/entry as a dictionary"""
        self.sanitize_tags()
        d = {}

        for column in self.__table__.columns:
            d[column.name] = getattr(self, column.name)

        return d

    @property
    def entry(self):
        """Returns the database entry for this object based on the primary key"""
        with self.session as session:
            entry = session.query(self.__class__).filter(self.primary_attribute() == self.primary)

        return entry

    @property
    def exists(self):
        """Returns True/False if the object exists in the database"""
        with self.session as session:
            count = (
                session.query(self.__class__)
                .filter(self.primary_attribute() == self.primary)
                .count()
            )

        return count != 0

    def sanitize_tags(self):
        if self.tagstring is None:
            log.debug(f"This object does not have tags: {self}")
            return

        if isinstance(self.tagstring, str):
            # If it starts as a string, split it into a list so we can sanitize tags individually
            self.tagstring = self.tagstring.split(",")

        if isinstance(self.tagstring, list):
            # Strip whitespace from each tag and join it back into a list
            self.tagstring = [tag.strip() for tag in self.tagstring]
            self.tagstring = ",".join(self.tagstring)

        if isinstance(self.tagstring, str):
            # Lowercase the entire string
            self.tagstring = self.tagstring.lower()
        else:
            # If we don't end up with a string, something went very wrong
            raise InvalidTagValueError(self.tagstring)

    @property
    def tags(self):
        breakpoint()
        self.sanitize_tags()
        return self.tagstring.split(",")

    @tags.setter
    def tags(self, value):
        breakpoint()
        self.tagstring = value
        self.sanitize_tags()

    def save(self):
        """Save or update the object to the database"""
        log.info(f"Saving {self}")
        self.sanitize_tags()

        try:
            with self.session as session:
                if self.exists:
                    log.debug(f"Updating: {self}")
                    entry = session.query(self.__class__).filter(
                        self.primary_attribute() == self.primary
                    )
                    entry.update(self.as_dict, synchronize_session=False)
                    session.commit()
                else:
                    log.info(f"New record: {self}")
                    session.add(self)
                    session.commit()
        except OperationalError as e:
            log.exception(f"Unable to save object {self}")
            raise DatabaseAccessError(f"Unable to save object {self} to {self.db}")

    @classmethod
    def primary_attribute(cls):
        """Stores the attribute reference for the primary key (useful for queries)"""
        return cls.__getattribute__(cls, cls.__primary__)

    @classmethod
    def get_key(cls, value):
        """Returns the database entry based on the given primary key"""

        with cls.db.session as session:
            return session.query(cls).get(value)


class DatabaseTable:
    db = dfdb
    orm = DatabaseEntry

    def __init__(self):
        log.debug(f"Created {self} -> {self.db}")

        # Generate tables for the entry class
        if issubclass(self.orm, EntryORM):
            self.db.generate_table(self.orm)

    @property
    def session(self):
        return DatabaseSession(self.db)

    def __repr__(self):
        return f"{self.__class__.__name__}[{self.orm.__name__}]"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.orm.__name__}]"

    def query(self):
        """Return a query for this table with an open session"""
        return self.session.session.query(self.orm)

    def primary(self, value):
        """Returns the database entry based on the given primary key"""
        return self.query().get(value)

    def all(self):
        """Returns a list of all objects"""
        return self.query().all()

    def keys(self):
        """Returns a list of all primary keys"""
        return [entry.primary for entry in self.all()]

    def keyed(self):
        """Returns a dictionary of all objects, keyed on the primary key"""
        return {entry.primary: entry for entry in self.all()}

    def filter(self, *args, **kwargs):
        """Filter the table and return the results on an open session"""
        return self.query().filter(*args, **kwargs)

    def tagged(self, *tag_list):
        """
        Filter the table using tags. Each entry in tag list is a string containing one or more
        tags. Tags may be comma separated and may be negated by prepending a dash. Entries must
        match at one or more of the tag lists specified.
        """
        if self.orm.tagstring is None:
            raise DatabaseEntryMissingColumnError(self.orm, "tagstring")

        matches = set()
        entries = self.query()

        for tags in tag_list:
            if isinstance(tag, str):
                tags = tags.split(",")

            for tag in tags:
                if tag[0] == "-":
                    tag = tag[1:]
                    entries = entries.filter(not self.orm.tagstring.contains(tag))
                else:
                    entries = entries.filter(self.orm.tagstring.contains(tag))

            matches.add(entries)
            entries = self.query()

        return matches



import re
import json
import time

import requests
import selenium
from bs4 import BeautifulSoup

from forgefrenzy.exceptions import WebResourceError
from forgefrenzy.blueplate.logger import log


class WebResource:
    base_url = "https://example.com"
    request_cooldown = 2
    last_request = time.time() - request_cooldown

    def __init__(self):
        raise RuntimeError("This class is not intended to be instantiated")

    @classmethod
    def get(cls, page=""):
        url = f"{cls.base_url}/{page}"
        delay = max((cls.last_request + cls.request_cooldown) - time.time(), 0.5)
        log.debug(f"Waiting {delay}s to avoid blacklisting")
        time.sleep(delay)
        cls.last_request = time.time()

        log.debug(f"requests.get({url})")

        try:
            return requests.get(url)
        except Exception as e:
            raise WebResourceError(f"Failed to GET {url}") from e

    @classmethod
    def json(cls, page=""):
        try:
            return cls.get(page).json()
        except Exception as e:
            raise WebResourceError(f"Failed to get JSON from {page}") from e

    @classmethod
    def soup(cls, page=""):
        try:
            return BeautifulSoup(cls.get(page).text, "html.parser")
        except Exception as e:
            raise WebResourceError(f"Failed to get BeautifulSoup from {page}") from e


class Webstore(WebResource):
    base_url = "https://dwarvenforge.com"
    product_url = "/products/{handle}.js"
    inventory_url = "/collections/all-products?page={page_number}"
    inventory_url_regex = "href=['\"]/collections/all-products/products/([^'\"]+)['\"]"
    inventory_count_regex = "([0-9]+) products"

    product_count = None
    product_handles = None
    products = []

    @classmethod
    def get_product_count(cls):
        response = cls.get(cls.inventory_url.format(page_number=1))
        product_count = re.search(cls.inventory_count_regex, response.text).group(1)
        return int(product_count)

    @classmethod
    def get_product_handles(cls, page_number=None):
        if page_number is not None:
            response = cls.get(cls.inventory_url.format(page_number=page_number))
            products_from_page = re.findall(cls.inventory_url_regex, response.text)
            return products_from_page

        product_count = cls.get_product_count()
        page_number = 0
        page_products = True
        all_products = []

        log.info(f"Refreshing product list, website reports {product_count} products")

        while page_products:
            page_number += 1
            page_products = cls.get_product_handles(page_number)
            all_products += page_products

            log.debug(f"Page {page_number}: {len(page_products)} products")

        log.info(f"Processed {page_number} pages of products: {len(all_products)} products")

        return all_products

    @classmethod
    def get_product(cls, handle):
        return Webstore.json(cls.product_url.format(handle=handle))

    @classmethod
    def refresh(cls):
        products = []
        for handle in cls.get_product_handles():
            products.append(cls.get_product(handle))

        return products


class Catalog(WebResource):
    base_url = "https://pieces.dwarvenforge.com"
    request_cooldown = 0.5

    @classmethod
    def get_data(cls):
        soup = cls.soup()
        data = soup.find(id="app").attrs["data-page"]

        try:
            as_json = json.loads(data)
        except Exception as e:
            raise CatalogResourceError(f"Failed to get JSON from Catalog inline page data") from e

        return as_json



import json
from datetime import datetime, timedelta
from pprint import pprint, pformat

from sqlalchemy import Column, Integer, String, DateTime, Boolean, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from forgefrenzy.db import DatabaseTable, DatabaseEntry, EntryORM
from forgefrenzy.store import Catalog, Webstore

from forgefrenzy.blueplate.logger import log


class Product(DatabaseEntry, EntryORM):
    __tablename__ = "product"
    __primary__ = "handle"

    default_dataset = "main"
    dataset = Column(String, default=default_dataset)
    handle = Column(String, primary_key=True)
    type = Column(String)
    tagstring = Column(String)
    sku = Column(String)
    name = Column(String)
    price = Column(Integer)
    available = Column(Boolean)
    quantity = Column(Integer)
    weight = Column(Integer)
    updated = Column(DateTime, server_default=func.now())

    def refresh(self, force=False):
        """Pull the latest data from the webstore and update the object"""
        if force or self.refreshable:
            webstore_data = Webstore.get_product(self.handle)
            as_dict = Product.from_webstore_to_dict(webstore_data)

            for key, value in as_dict.items():
                setattr(self, key, value)

            self.updated = datetime.now()
            self.save()

    def save(self):
        if self.updated is None:
            self.updated = datetime.now()

        super().save()

    @property
    def set(self):
        sku = self.sku
        if sku.endswith("-U") or sku.endswith("-P"):
            sku = sku[:-3]

        return Set.primary(sku)

    @property
    def refreshable(self):
        try:
            return self.next_update <= datetime.now()
        except:
            return True

    @property
    def next_update(self):
        if self.updated is None:
            # If the product has never updated, return 0 to force update
            return 0

        refresh_frequency = 60 * 24 * 7

        if self.type.lower() in ["dwarvenite", "resin"]:
            refresh_frequency = 60 * 24

            if self.available and self.quantity > 0:
                if self.quantity < 10:
                    refresh_frequency = 1
                elif self.quantity < 25:
                    refresh_frequency = 5
                elif self.quantity < 50:
                    refresh_frequency = 10
                elif self.quantity < 100:
                    refresh_frequency = 30
                elif self.quantity < 200:
                    refresh_frequency = 60

        return self.updated + timedelta(minutes=refresh_frequency)

    @classmethod
    def from_webstore_to_dict(cls, webstore_data):
        """
        Takes the JSON blob returned by the webstore and converts it into a dictionary matching
        the structure of the ORM object
        """
        as_dict = {}
        as_dict["handle"] = webstore_data["handle"]
        as_dict["type"] = webstore_data["type"]
        as_dict["tagstring"] = webstore_data["tags"]

        if len(webstore_data.get("variants", [])) == 0:
            log.warning(
                "No variant information is available for '{self.title}', creating entry with partial data"
            )
            as_dict["sku"] = webstore_data.get("sku")
            as_dict["name"] = webstore_data.get("title")
            as_dict["price"] = webstore_data.get("price")
            as_dict["available"] = webstore_data.get("available")
        else:
            if len(webstore_data["variants"]) > 1:
                log.warning(
                    "More than one variant is available for '{self.title}', creating entry with first variant"
                )

            variant = webstore_data["variants"][0]

            as_dict["sku"] = variant.get("sku")
            as_dict["name"] = variant.get("name")
            as_dict["price"] = variant.get("price")
            as_dict["available"] = variant.get("available")
            as_dict["quantity"] = variant.get("inventory_quantity")
            as_dict["weight"] = variant.get("weight")

        return as_dict

    @classmethod
    def from_webstore(cls, webstore_data):
        return cls(**cls.from_webstore_to_dict(webstore_data))

    @classmethod
    def from_handle(cls, handle):
        return Product.from_webstore(Webstore.get_product(handle))


class Products(DatabaseTable):
    orm = Product

    def refresh(self, force=False):
        handles = self.refresh_handles(force)
        products = self.refresh_products(handles, force)
        return products

    def refresh_handles(self, force=False):
        if force:
            return Webstore.get_product_handles()

    def refresh_products(self, handles, force=False):
        """
        Refresh the product information for the given handles.
        If the handle does not exist in the database, data will be fetched from the webstore and added to the database.
        If force is enabled, all entries in the database will be refreshed.
        If neither, the object will determine whether it is ready to be refreshed based on the refreshable attribute.
        :param handles:
        :type handles:
        :param force:
        :type force:
        :return:
        :rtype:
        """
        handles = handles or self.keys()
        products = []

        for handle in handles:
            log.debug(f"Checking if refreshable: {handle}")
            product = self.primary(handle)

            if product is None:
                log.debug(f"No record for {handle}, creating a new record")
                # Product is new
                product = Product.from_handle(handle)
                product.save()
                products.append(product)
            elif force or product.refreshable:
                if force:
                    log.info(f"Forcing refresh of product: {product}")
                else:
                    log.debug(f"Refreshing product {product} due to timer")

                product.refresh(force)
                products.append(product)
            else:
                log.info(f"Skipping product because it is not eligible for a refresh: {product}")

        return products

    def painted(self):
        return self.tagged("painted")

    def unpainted(self):
        return self.tagged("unpainted")


class Set(DatabaseEntry, EntryORM):
    __tablename__ = "set"
    __primary__ = "sku"
    default_dataset = "main"
    dataset = Column(String, default=default_dataset)
    sku = Column(String, primary_key=True)
    name = Column(String)
    img = Column(String)
    tagstring = Column(String)


class Sets(DatabaseTable):
    orm = Set
    remap = {
        "sku": "value",
        "name": "text",
        "img": "img",
        "tagstring": "tag",
    }

    def refresh(self):
        """
        Contacts the catalog to pull down the latest sets list.
        Injests them into Set objects and saves them to the database.
        Returns the list of Set objects for convenience.
        """
        catalog_data = Catalog.get_data()
        log.debug(f"Found {len(json.dumps(catalog_data))} bytes in catalog data")
        entries = catalog_data["props"]["data_hints"]
        log.debug(f"Found {len(entries)} total entries in catalog data")
        set_entries = [entry for entry in entries if "set_sku" not in entry.keys()]
        log.debug(f"Found {len(set_entries)} set entries in catalog data")

        sets = []
        for set_entry in set_entries:
            remapped = {k: set_entry[v] for k, v in self.remap.items()}
            new_set = Set(**remapped)
            sets.append(new_set)
            new_set.save()

        return sets


class PartList(DatabaseEntry, EntryORM):
    __tablename__ = "partlist"
    __primary__ = "id"
    default_dataset = "main"
    id = Column(Integer, primary_key=True)
    set = Column(String, ForeignKey("set.sku"))
    piece = Column(String, ForeignKey("pieces.sku"))

    quantity = Column(Integer)

    def __repr__(self):
        return f"{self.__class__.__name__} <{self.quantity} {self.piece} in {self.set}>"

    def __str__(self):
        return f"{self.__class__.__name__} <{self.quantity} {self.piece} in {self.set}>"


class PartLists(DatabaseTable):
    orm = PartList
    remap = {
        "set": {"set": "value", "piece": "piece_sku", "quantity": "piece_quantity"},
        "piece": {"piece": "value", "set": "set_sku", "quantity": "piece_quantity"},
    }

    def refresh(self):
        """
        Contacts the catalog to pull down the latest sets list.
        Reads the quantity of pieces
        Returns the list of PartList objects for convenience.
        """
        catalog_data = Catalog.get_data()
        log.debug(f"Found {len(json.dumps(catalog_data))} bytes in catalog data")
        entries = catalog_data["props"]["data_hints"]
        log.debug(f"Found {len(entries)} total entries in catalog data")
        set_entries = [entry for entry in entries if "set_sku" not in entry.keys()]
        log.debug(f"Found {len(set_entries)} set entries in catalog data")
        piece_entries = [entry for entry in entries if "piece_sku" not in entry.keys()]
        log.debug(f"Found {len(piece_entries)} piece entries in catalog data")

        partlists = []
        for piece in piece_entries:
            remapped = {k: piece[v] for k, v in self.remap["piece"].items()}
            new_partlist = PartList(**remapped)
            partlists.append(new_partlist)
            new_partlist.save()

        return partlists


class Piece(DatabaseEntry, EntryORM):
    __tablename__ = "pieces"
    __primary__ = "sku"
    default_dataset = "main"
    dataset = Column(String, default=default_dataset)
    sku = Column(String, primary_key=True)
    name = Column(String)
    img = Column(String)
    tagstring = Column(String)


class Pieces(DatabaseTable):
    orm = Piece
    remap = {
        "sku": "value",
        "name": "text",
        "img": "img",
        "tagstring": "piece_tag",
    }

    def refresh(self):
        """
        Contacts the catalog to pull down the latest pieces list.
        Injests them into Piece objects and saves them to the database.
        Returns the list of Piece objects for convenience.
        """
        catalog_data = Catalog.get_data()
        log.debug(f"Found {len(json.dumps(catalog_data))} bytes in catalog data")
        entries = catalog_data["props"]["data_hints"]
        log.debug(f"Found {len(entries)} total entries in catalog data")
        piece_entries = [entry for entry in entries if "piece_sku" not in entry.keys()]
        log.debug(f"Found {len(piece_entries)} piece entries in catalog data")

        pieces = []
        for piece in piece_entries:
            remapped = {k: piece[v] for k, v in self.remap.items()}
            new_piece = Piece(**remapped)
            pieces.append(new_piece)
            new_piece.save()

        return pieces



from forgefrenzy.products import Product, Products, Set, Sets, PartList, PartLists, Piece, Pieces
from forgefrenzy.store import Webstore, Catalog

__all__ = [
    "Product",
    "Products",
    "Set",
    "Sets",
    "PartList",
    "PartLists",
    "Piece",
    "Pieces",
    "Catalog",
    "Webstore",
]



""" 
blueplate.special: a blueprint for sensible command line tools
"""

import os
import sys
import argparse
import inspect

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from .imports import minimock, mock, doctest
from .version import version
from .logger import log


class Special:
    """blueplate.special: a blueprint for sensible command line tools"""

    default_level = log.config.default
    options = []
    errors = {}
    testmode = False

    def __init__(self, argv=None):
        """
        >>> Special().argv == sys.argv[1:]
        True
        >>> Special("--debug").argv == ["--debug"]
        True
        >>> Special("--debug --log").argv == ["--debug", "--log"]
        True
        >>> Special(["--debug", "--log --debugv"]).argv
        ['--debug', '--log', '--debugv']
        """
        self.__module = sys.modules[self.__class__.__module__]
        self.__version = self.__module.version
        self.description = self.__doc__.splitlines()[0].strip()
        if ":" in self.description:
            self.name, self.description = self.description.split(":", 1)
            self.name = self.name.strip()
            self.description = self.description.strip()
        else:
            self.name = self.__class__.__name__

        log.critical(f"Creating object with args {argv=}")
        if argv is not None:
            if isinstance(argv, str):
                log.debug(
                    f"Creating {self.__class__.__name__} object using provided string {argv=}"
                )
                self.argv = argv.split(" ")
            elif isinstance(argv, list):
                log.debug(f"Creating {self.__class__.__name__} object using provided list {argv=}")
                self.argv = argv
            else:
                log.warning(
                    "Provided value for argv is neither a string nor list, continuing without args"
                )
                self.argv = []
        else:
            log.debug(f"Creating {self.__class__.__name__} object using {sys.argv[1:]=}")
            self.argv = sys.argv[1:]

        # Remove empty strings
        self.argv = [arg for subarg in self.argv for arg in subarg.split(" ") if arg]

        self.__options = None
        self.__args = None
        self.__argparser = None
        self.__errors = None

    def main(self):
        """Stub main
        >>> sys.argv = "special.py --log".split()
        >>> special = Special("")
        >>> mock("special.help")
        >>> special.main()
        Called special.help()
        """
        log.config.update(self)

        if self.args.test:
            self.test()
            return

        if self.args.version:
            print(self.version)
            return

        self.fn()

    @property
    def version(self):
        return self.__version

    def fn(self):
        self.help()

    def help(self):
        """Show usage text
        # >>> special.argv
        >>> special.help()
        usage: ...
        ...
        options:
        ...
        arguments:
        ...
        error codes:
        ...
        """
        print(
            f"""\
{self.show_usage()}
{self.show_args()}

{self.show_errors()}
"""
        )

    @property
    def argparser(self):
        """Generate an arg parser object with all the args from the caller and the util module
        >>> special.argparser
        ArgumentParser(...
        >>> special.argparser.description == special.description
        True
        """
        if self.__argparser is not None:
            return self.__argparser

        self.__argparser = argparse.ArgumentParser(
            prog=self.name,
            description=self.description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        # Add all arguments to the parser
        for option in self.options:
            option = option.copy()
            args = option.pop("args")
            kwargs = option
            self.__argparser.add_argument(*args, **kwargs)

        return self.__argparser

    @property
    def args(self):
        """Get the args, handle default args
        # >>> print(Special("--debug").show_args())
        >>> Special("--debug").args.debug
        True
        >>> Special("--debug").args.test
        False
        """
        if self.__args is not None:
            return self.__args

        try:
            self.__args = self.argparser.parse_args(self.argv)
        except Exception as ex:
            self.error("bad-args", e=ex)

        return self.__args

    def show_usage(self):
        return self.argparser.format_help()

    def show_args(self):
        """Display a list of all arguments parsed
        >>> print(Special("--log --debugv").show_args())
        arguments:
          log               True
          console           False
          debug             False
          debugv            True
          test              False
          version           False
        """
        output = ["arguments:"]

        for arg, val in vars(self.args).items():
            output.append(f"  {arg:17} {val}")

        return "\n".join(output)

    def show_errors(self):
        """Render a table of error messages and exit codes
        >>> print(special.show_errors())
        error codes:
          128               The error key {key} is undefined.
        ...
        """
        output = ["error codes:"]

        for error, index in self.errors.values():
            output.append(f"  {index:<17} {error}")

        return "\n".join(output)

    @property
    def config(self):
        defaults_file = os.path.join(os.path.dirname(__file__), "defaults.toml")
        config_file = os.path.join(
            os.path.dirname(self.__module.__file__), "..", "..", "pyproject.toml"
        )

        with open(defaults_file, "rb") as f:
            defaults = tomllib.load(f).get("tools", {}).get("blueplate", {})

        try:
            with open(config_file, "rb") as f:
                config = tomllib.load(f).get("tools", {}).get("blueplate", {})
        except FileNotFoundError:
            log.warning(f"Could not find package configuration TOML: {config_file}")
            config = {}

        return {
            "options": defaults.get("options", {}) | config.get("options", {}),
            "errors": defaults.get("errors", {}) | config.get("errors", {}),
        }

    @property
    def options(self):
        if self.__options is not None:
            return self.__options

        self.__options = []

        for arg, details in self.config["options"].items():
            details["args"] = [f"--{arg}"] + details.pop("alternates", [])
            self.__options.append(details)

        return self.__options

    @property
    def errors(self):
        """
        >>> special.errors["bad-args"]
        ('Arguments provided were invalid: {e}', 129)
        >>> special.errors["test-error"]
        ('DocTest exectution failed due to exception during execution', 131)
        """
        if self.__errors is not None:
            return self.__errors

        self.__errors = {}

        for error, message in self.config["errors"].items():
            error_code = list(self.config["errors"].keys()).index(error) + 128
            self.__errors[error] = (message, error_code)

        return self.__errors

    def error(self, key, **kwargs):
        """Format the error string, raise an exception or log the error and exit with the error code
        >>> special.error("not-a-real-error")
        Traceback (most recent call last):
        SystemExit: 128
        >>> special.error("unknown-error", key="missing-key")
        Traceback (most recent call last):
        ...
        TypeError: ...error() got multiple values for argument 'key'
        >>> special.error("test-failure")
        Traceback (most recent call last):
        SystemExit: 130
        >>> special.error("test-error")
        Traceback (most recent call last):
        SystemExit: 131
        """
        if key not in self.errors.keys():
            # If not defined, use catchall
            kwargs["key"] = key
            key = "unknown-error"

        string, code = self.errors[key]
        formatted_string = string.format(**kwargs)

        log.critical(f"Exit {code}: {formatted_string}")
        print(f"Exit {code}: {formatted_string}")

        exit(code)

    def test_globs(self):
        """Define test tools for Doctest
        >>> list(special.test_globs().keys())
        ['mock', 'minimock', 'special']
        """
        from .imports import minimock, mock

        globs = {"mock": mock, "minimock": minimock, "special": type(self)("")}
        return globs

    def test(self, mock_test=False):
        """Run Doctests
        >>> special.test(mock_test=True)
        Called doctest.testmod(
            <module '...' from '.../special.py'>,
        ...
        """
        module = sys.modules[self.__class__.__module__]

        result = test_module(module, self.test_globs(), mock_test)

        if result is not None:
            self.error(result)


def test_module(module, test_globs=None, mock_test=False):
    """Run Doctests
    >>> test_module(sys.modules[special.__module__], mock_test=True)
    Called doctest.testmod(
        <module '...' from '.../special.py'>,
    ...
    """

    filename = os.path.basename(module.__file__)
    testmod = doctest.testmod

    if mock_test:
        doctest.testmod = minimock.Mock("doctest.testmod")
        log.alert(f"Simulating doctests of module {module} in file {filename} with mock")
        log.debugv(f"{doctest.testmod=}")
    else:
        log.alert(f"Running doctests of module {module} in file {filename}")
        log.debugv(f"{doctest.testmod=}")

    try:
        log.critical(f"{module=}")
        doctest.testmod(
            module,
            raise_on_error=True,
            verbose=True,
            extraglobs=test_globs,
            optionflags=doctest.ELLIPSIS,
        )
    except doctest.DocTestFailure as e:
        log.critical(f"doctest failed: DocTestFailure")
        log.config.console = True
        log.config.level = "ERROR"
        try:
            e_path = f"{os.path.basename(e.test.filename)}:{(e.test.lineno + e.example.lineno + 1)}"
        except:
            e_path = "Unknown"

        e_str = f"""
=======================================
    DocTestFailure
        {e.test.name}
        {e_path}
=======================================
        
    Trying:
        {e.example.source}
    Expected:
        {e.example.want.strip() or "-----"}
        
    Got:
        {e.got}

=======================================
"""
        log.error(e_str)
        return "test-failure"
    except doctest.UnexpectedException as e:
        log.critical(f"doctest failed: UnexpectedException")
        log.config.console = True
        log.config.level = "ERROR"
        try:
            e_path1 = (
                f"{os.path.basename(e.test.filename)}:{(e.test.lineno + e.example.lineno + 1)}"
            )
        except:
            e_path1 = "Unknown"
        try:
            e_path2 = f"...{os.path.dirname(e.exc_info[2].tb_next.tb_next.tb_frame.f_code.co_filename)[-15:]}/{os.path.basename(e.exc_info[2].tb_next.tb_next.tb_frame.f_code.co_filename)}:{e.exc_info[2].tb_next.tb_next.tb_frame.f_lineno} in '{e.exc_info[2].tb_next.tb_next.tb_frame.f_code.co_name}'"
        except:
            e_path2 = "Unknown"
        e_str = f"""
=======================================
    UnexpectedException
        {e.test.name}
        {e_path1}
=======================================
        
    Trying:
        {e.example.source}
    Expected:
        {e.example.want.strip() or "-----"}
        
    Exception:
        {e.exc_info[0].__name__}: {e.exc_info[1]}
        {e_path2}

=======================================

"""
        log.exception(e_str)
        return "test-error"
    finally:
        doctest.testmod = testmod


def get_caller_module(level=None, ignore_direct_caller=True, ignore_self=True):
    """Return the module which called this function
    ignore_direct_caller    Starts at the caller of the caller
    ignore_self             Ignore all this of this module
    """
    ignore_modules = []

    if ignore_self:
        ignore_modules.append(sys.modules[__name__])

    frame = inspect.currentframe()

    from pprint import pprint

    # pprint(frame)
    # print(dir(frame))
    index = 0
    stack = []
    while frame.f_back is not None:
        # print(f"#{index} {frame=}")
        stack.append(frame)
        index += 1
        frame = frame.f_back

        if level is not None and level >= index:
            break

    for k, v in sys.modules.items():
        if v == k:
            return sys.modules[k]

    print(dir(frame))
    # help(frame)
    print({k: getattr(frame, k) for k in dir(frame)})
    # pprint(sys.modules)

    if __name__ == "__main__":
        module_obj = sys.modules[__name__]
    else:
        module_name = inspect.getmodulename(inspect.stack()[0][1])
        log.debug(f"Found root module '{module_name}'")
        module_obj = sys.modules[module_name]

        # print(f"{__name__=}")
        module_obj = sys.modules["__main__"]
        # print(f"{module_obj=}")
        pprint({k: getattr(module_obj, k) for k in dir(module_obj)})
    return module_obj


if __name__ == "__main__":
    script = Special()
    script.main()



""" This is a util function containing common patterns for scripts
>>> isinstance(log, CustomLogger)
True
>>> logging.OUTPUT
8
>>> log.output
<function CustomLogger.add_custom_levels.<locals>.custom_log_fn_wrapper.<locals>.custom_log_fn at 0x...>
>>> logging.DEBUGV
9
>>> log.debugv
<function CustomLogger.add_custom_levels.<locals>.custom_log_fn_wrapper.<locals>.custom_log_fn at 0x...>
>>> logging.ALERT
51
>>> log.alert
<function CustomLogger.add_custom_levels.<locals>.custom_log_fn_wrapper.<locals>.custom_log_fn at 0x...>
"""

import math
import builtins
import logging

from .imports import minimock, mock


class LoggerConfig:
    __default = logging.ERROR
    long_formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d %(levelname)-10.10s [%(name)s.%(module)s.%(funcName)s:%(lineno)d] %(message)s"
    )
    short_formatter = logging.Formatter("[%(levelname)-8.8s]  %(message)s")

    def __init__(self, log, default=None):
        self.log = log
        self.log.handlers = []
        self.name = self.log.name
        self.default = default or self.__default
        self.__level = self.default
        self.log.setLevel(0)

        self.__file_handler = None
        self.__console_handler = None

        self.null_handler = logging.NullHandler()
        self.log.addHandler(self.null_handler)

    @property
    def file_handler(self):
        if self.__file_handler is not None:
            return self.__file_handler

        self.__file_handler = logging.FileHandler(f"./{self.name}.log", encoding="utf8")
        self.__file_handler.setFormatter(self.long_formatter)
        self.__file_handler.setLevel(0)
        return self.__file_handler

    @property
    def console_handler(self):
        if self.__console_handler is not None:
            return self.__console_handler

        self.__console_handler = logging.StreamHandler()
        self.__console_handler.setFormatter(self.short_formatter)
        self.__console_handler.setLevel(self.level)
        return self.__console_handler

    def update(self, script):
        """Take an argparser object and look for logging related args to update the config"""
        # Enable console logging if any verbose options are selected
        if script.args.debugv or script.args.debug or script.args.console:
            self.console = True

        # Set a new default level if present in the script object
        try:
            self.default = script.default_level
        except AttributeError:
            pass

        if script.args.debugv:
            level = "DEBUGV"
        elif script.args.debug:
            level = "DEBUG"
        else:
            level = self.default

        self.level = level

        if script.args.log:
            log.config.file = True
        else:
            log.warning("No file logging enabled")

    @property
    def next(self):
        """Get next standard log level up
        >>> log.config.level = 9
        >>> log.config.level
        9
        >>> log.config.next
        10
        >>> log.config.level = 10
        >>> log.config.next
        20
        >>> log.config.level = 50
        >>> log.config.next
        50
        >>> log.config.level = 51
        >>> log.config.next
        51
        """
        next_level = math.ceil((self.level + 1) / 10) * 10

        # Get at least logging.INFO
        next_level = max(next_level, logging.DEBUG)
        # Get at most logging.CRITICAL
        next_level = min(next_level, logging.CRITICAL)
        # If log level is set higher than logging.CRITICAL, use that
        next_level = max(next_level, self.level)
        return next_level

    @property
    def effective(self):
        return self.log.getEffectiveLevel()

    @property
    def level(self):
        return self.__level

    @level.setter
    def level(self, level):
        if self.level != level:
            if isinstance(level, str):
                level = getattr(logging, level.upper())
            self.__level = level

            if self.__console_handler is not None:
                self.console_handler.setLevel(level)

            self.log.log(self.level, f"Log level set to {level}")

    def reset_level(self):
        self.level = self.default
        self.log.debug(f"Disabled debug debugging, reset to {self.default}")

    @property
    def console(self):
        return bool([h for h in self.log.handlers if not isinstance(h, logging.StreamHandler)])

    @console.setter
    def console(self, enable):
        if enable:
            self.log.addHandler(self.console_handler)
            self.log.alert(f"Enabled console logging at {logging.getLevelName(self.level)}")
        else:
            self.log.debug(f"Disabled console logging")
            self.log.handlers = [
                h for h in self.log.handlers if not isinstance(h, logging.StreamHandler)
            ]

    @property
    def file(self):
        return bool([h for h in self.log.handlers if not isinstance(h, logging.FileHandler)])

    @file.setter
    def file(self, enable):
        if enable:
            self.log.addHandler(self.file_handler)
            self.log.alert(
                f"Enabled file logging at {logging.getLevelName(self.file_handler.level)}"
            )
        else:
            self.log.handlers = [
                h for h in self.log.handlers if not isinstance(h, logging.FileHandler)
            ]


class CustomLogger(logging.getLoggerClass()):
    __config = None
    # Configure custom log levels
    custom_levels = [
        {"level": 51, "name": "alert"},
        {"level": 9, "name": "debugv"},
        {"level": 8, "name": "output"},
    ]

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)
        self.config
        self.add_custom_levels()

    def add_custom_levels(self):
        for custom in self.custom_levels:

            def custom_log_fn_wrapper(level):
                """Generate a logger function for the given level
                >>> output = custom_log_fn_wrapper(52)
                >>> output
                <function custom_log_fn_wrapper.<locals>.custom_log_fn at 0x...>
                >>> log_dot_log = log.log
                >>> log.log = mock('log.log')
                >>> log.setLevel(logging.DEBUGV)
                >>> output("Test")
                Called log.log(8, 'Test')
                >>> log.log = log_dot_log
                """

                def custom_log_fn(message, *args, **kws):
                    # args = args or None
                    if self.isEnabledFor(level):
                        # Yes, logger takes its '*args' as 'args'.
                        self.log(level, message, *args, **kws)

                    if level == logging.OUTPUT:
                        builtins.print(message)

                return custom_log_fn

            logging.addLevelName(custom["level"], custom["name"].upper())
            setattr(self, custom["name"], custom_log_fn_wrapper(custom["level"]))
            setattr(logging, custom["name"].upper(), custom["level"])

    @property
    def config(self):
        if self.__config is not None:
            return self.__config

        self.__config = LoggerConfig(self)

        return self.__config


logging.setLoggerClass(CustomLogger)
logging.root = CustomLogger("root")
log = logging.getLogger()



""" Command line access using Blueplate Special
>>> m = Main("")
>>> isinstance(m, Main)
True
"""
import json

from forgefrenzy.blueplate.special import Special
from forgefrenzy.blueplate.version import version
import forgefrenzy.blueplate.logger

from forgefrenzy import *

log = forgefrenzy.blueplate.logger.log


class Main(Special):
    """forgefrenzy: A collection of tools to nurture your addiction to Dwarven Forge
    >>> isinstance(script, Main)
    True
    >>> isinstance(script, Special)
    True
    >>> issubclass(script.__class__, Special)
    True
    """

    def fn(self):
        pieces = Pieces()
        sets = Sets()
        partlists = PartLists()
        products = Products()

        # pieces.refresh()
        # sets.refresh()
        # partlists.refresh()
        # products.refresh()

        product = products.all()[100]
        set = product.set
        breakpoint()

    def test(self, mock_test=False):
        """Run local tests and tests on other mods
        >>> script.test(True)
        Called doctest.testmod(
            <module '...__main__' from '.../__main__.py'>,
        ...
        Called doctest.testmod(
            <module '...blueplate.special' from '.../blueplate/special.py'>,
        ...
        Called doctest.testmod(
            <module '....blueplate.logger' from '.../blueplate/logger.py'>,
        ...
        """
        from .blueplate import logger, special

        super().test(mock_test)
        response = special.Special().test(mock_test)
        special.test_module(logger, mock_test=mock_test)


if __name__ == "__main__":
    script = Main()
    script.main()
